from ast import Not
import numpy as np
import numpy.typing as npt
from typing import Literal
from scipy.stats import rv_continuous
from scipy._lib.doccer import inherit_docstring_from
from scipy.stats._multivariate import multi_rv_generic
from scipy.stats._distn_infrastructure import _ShapeInfo

from .pdf import generate_alpha_stable_pdf
from .random import sample_alpha_stable_vector, IsotropicSampler, UnivariateSampler, EllipticSampler, DiscreteSampler, BaseSpectralMeasureSampler
from .random.cms_univariate_sampler import sample_cms
from .random.alpha_stable_sampler import estimate_number_of_convergence_terms
import logging

class alpha_stable_gen(rv_continuous):
    _parameterization: Literal["S1", "S0"] = "S1"

    @property
    def parameterization(self):
        return self._parameterization

    @parameterization.setter
    def parameterization(self, parametrization: Literal["S1", "S0"]):
        if parametrization not in ("S1", "S0"):
            raise ValueError(f"Parametrization '{parametrization}' not supported. Use 'S1' (default) or 'S0'.")
        self._parameterization = parametrization

    def with_parametrization(self, parametrization: Literal["S1", "S0"]):
        self.parameterization = parametrization
        return self

    def _shape_info(self):
        return [_ShapeInfo("alpha", False, (0, 2), (False, True)),
                _ShapeInfo("beta", False, (-1, 1), (True, True))]

    @inherit_docstring_from(rv_continuous)
    def pdf(self, x, *args, **kwds):
        # override base class version to correct location for S1 parameterization
        (alpha, beta), delta, gamma = self._parse_args(*args, **kwds)
        (x, alpha, beta, delta, gamma), size = _sanitize_array_args(x, alpha, beta, delta, gamma)
        if size == (0,):
            return np.array([])

        if np.all(alpha == alpha[0]) \
        and np.all(beta == beta[0]) \
        and np.all(delta == delta[0]) \
        and np.all(gamma == gamma[0]) \
        and np.all(np.diff(x) > 0):
            alpha_, beta_, delta_, gamma_ = alpha[0], beta[0], delta[0], gamma[0]
            return super().pdf(
                x,
                alpha_,
                beta_,
                loc = self._get_shift_term(alpha_, beta_, gamma_, delta_, self.parameterization),
                scale = gamma_
            ).reshape(size)
        else:
            pdf = [
                super(self.__class__, self).pdf(
                    x_,
                    alpha_,
                    beta_,
                    loc = self._get_shift_term(alpha_, beta_, gamma_, delta_, self.parameterization),
                    scale = gamma_
                ) for x_, alpha_, beta_, delta_, gamma_ in zip(x, alpha, beta, delta, gamma)
            ]
            return np.asarray(pdf).reshape(size)

    def _pdf(self, x, alpha, beta):
        (x, alpha, beta), size = _sanitize_array_args(x, alpha, beta)
        if size == (0,):
            return np.array([])

        pdf: np.ndarray
        if np.all(alpha == alpha[0]) and np.all(beta == beta[0]) and np.all(np.diff(x)>0):
            # Note: -1/2 and 1/2 have been chosen as arbitrary "close" values to x for interpolation to work.
            padded_x = np.concatenate([[x[0] - 1/2], x, [x[-1] + 1/2]])
            padded_pdf = generate_alpha_stable_pdf(padded_x, alpha[0], beta[0], 1, 0)
            pdf = padded_pdf[1:-1]

        else:
            pdf = np.array([
                generate_alpha_stable_pdf([xi - 1/2, xi, xi + 1/2], ai, bi, 1, 0)[1]
                for xi, ai, bi in zip(x, alpha, beta)
            ])

        return pdf.reshape(size)

    def _argcheck(self, alpha, beta):
        return (0 < alpha) & (alpha <= 2) & (-1 <= beta) & (beta <= 1)

    @staticmethod
    def _vectorized_rvs(alpha, beta, gamma, delta, parametrization, size = None, random_state = None):
        if size is None:
            (alpha, beta), size = _sanitize_array_args(alpha, beta)
        else:
            alpha = np.broadcast_to(alpha, size).ravel()
            beta = np.broadcast_to(beta, size).ravel()
            gamma = np.broadcast_to(gamma, size).ravel()
            delta = np.broadcast_to(delta, size).ravel()

        shift: np.ndarray
        if parametrization == "S0":
            shift = np.where(
                alpha == 1,
                delta,
                delta - gamma * beta * np.tan(alpha * np.pi / 2)
            )
        elif parametrization == "S1":
            shift = np.where(
                alpha == 1,
                delta + beta * 2 / np.pi * gamma * np.log(gamma),
                delta
            )

        if size == (0,):
            return np.array([])

        number_of_samples = int(np.prod(size))

        params = np.column_stack((alpha, beta, gamma, shift))
        unique_params, inverse = np.unique(params, axis=0, return_inverse=True)

        out = np.empty(number_of_samples)

        for idx, (alpha_, beta_, gamma_, shift_) in enumerate(unique_params):
            mask = inverse == idx
            number_of_samples_for_this_pair = mask.sum()

            samples = alpha_stable_gen._single_valued_rvs(alpha_, beta_, gamma_, shift_, number_of_samples_for_this_pair, random_state)

            # Assign samples to all rows that share this (alpha, beta) pair
            out[(inverse == idx)] = samples

        return out.reshape(size)
    

    @staticmethod
    def _single_valued_rvs(alpha_, beta_, gamma_, shift_, size, random_state):
        # patch: always use CMS sampler for now
        if True or alpha_ >= 1 and beta_ != 0 or estimate_number_of_convergence_terms(0.01, alpha_, gamma_**alpha_) >= 50000:
            logging.debug(f"Using CMS sampler for alpha = {alpha_}, beta = {beta_}")
            samples = sample_cms(alpha_, beta_, size, random_state)
            return samples * gamma_ + shift_
        else:
            logging.debug(f"Using Univariate sampler for alpha = {alpha_}, beta = {beta_}")
            sampler = UnivariateSampler(alpha_, beta_, gamma=gamma_)
            samples = sample_alpha_stable_vector(
                alpha_,
                sampler,
                size,
                shift_,
                random_state = random_state,
            )
        return samples


    def _rvs(self, alpha, beta, size=None, random_state=None):
        return self._vectorized_rvs(alpha, beta, 1, 0, size = size, random_state = random_state)


    @inherit_docstring_from(rv_continuous)
    def rvs(self, *args, **kwds):

        kwds.pop("discrete", None)
        random_state = kwds.pop("random_state", None)
        (alpha, beta), delta, gamma, size = self._parse_args_rvs(*args, **kwds)

        return self._vectorized_rvs(alpha, beta, gamma, delta, self.parameterization, size=size, random_state=random_state)


    def _get_shift_term(self, alpha: float, beta: float, gamma: float, delta: float, parameterization: Literal["S0", "S1"] | None = None) -> float:
        parameterization = parameterization or self.parameterization
        if parameterization == "S0":
            return delta
        elif parameterization == "S1":
            if alpha != 1:
                return delta
            else:
                return delta + (2 / np.pi) * beta * gamma * np.log(gamma)
        else:
            raise AssertionError("Unknown parametrization type")


class multivariate_alpha_stable_gen(multi_rv_generic):
    def rvs(self,
        alpha: float,
        spectral_measure_sampler: BaseSpectralMeasureSampler | Literal["standard_isotropic_2d", "standard_isotropic_3d", "1x2_elliptic_2d", "1x2x4_elliptic_3d", "coin_flip_discrete"] = "standard_isotropic_2d",
        shift: np.ndarray = 0,
        size: int | None = None,
        random_state: None | int | np.random.RandomState | np.random.Generator = None,
    ):
        spectral_measure_sampler = self._check_spectral_measure_sampler(alpha, spectral_measure_sampler)
        assert isinstance(spectral_measure_sampler, BaseSpectralMeasureSampler)
        samples = sample_alpha_stable_vector(alpha, spectral_measure_sampler, size or 1, shift, random_state = random_state)
        if size is None:
            return samples[0]
        return samples

    def _check_spectral_measure_sampler(self, alpha: float, spectral_measure_sampler: BaseSpectralMeasureSampler | str) -> BaseSpectralMeasureSampler:
        if isinstance(spectral_measure_sampler, BaseSpectralMeasureSampler):
            return spectral_measure_sampler

        if spectral_measure_sampler == "standard_isotropic_2d":
            return IsotropicSampler(2, alpha, 1)

        if spectral_measure_sampler == "standard_isotropic_3d":
            return IsotropicSampler(3, alpha, 1)

        if spectral_measure_sampler == "1x2_elliptic_2d":
            return EllipticSampler(2, alpha, sigma=[[1, 0], [0, 2]])

        if spectral_measure_sampler == "1x2x4_elliptic_3d":
            return EllipticSampler(3, alpha, sigma=[[1, 0, 0], [0, 2, 0], [0, 0, 4]])

        if spectral_measure_sampler == "coin_flip_discrete":
            return DiscreteSampler(alpha, [-1, 1], [0.5, 0.5])

        raise ValueError("Unknown spectral measure sampler")

def _sanitize_array_args(*args):
    assert len(args) > 0
    args = list(map(np.asarray, args))
    args = np.broadcast_arrays(*args)
    size = args[0].shape
    args = list(map(np.ravel, args))
    return args, size



import numpy as np
from .util import get_random_state_generator

def sample_cms(alpha : float, beta : float, size : int = 1, random_state : None | int | np.random.RandomState | np.random.Generator = None) -> np.ndarray:
    """
    Generate alpha-stable random variables in Nolan Type 1 parametrization
    with gamma = 1 and delta = 0.

    Parameters
    ----------
    alpha : float in (0,2]
    beta  : float in [-1,1]
    size  : int
    random_state : int or None

    Returns
    -------
    samples : ndarray
        Samples from S(alpha, beta, 1, 0; 1)
    """

    if not (0 < alpha <= 2):
        raise ValueError("alpha must be in (0,2]")
    if not (-1 <= beta <= 1):
        raise ValueError("beta must be in [-1,1]")
    random_state = get_random_state_generator(random_state)

    # CMS base variables
    U = random_state.uniform(-np.pi/2, np.pi/2, size)
    W = random_state.exponential(1.0, size)

    # Small tolerance for alpha=1 detection
    tol = 1e-12

    if abs(alpha - 1.0) > tol:
        # -------- alpha != 1 case --------

        # theta0 = (1/alpha) arctan(beta tan(pi alpha / 2))
        theta0 = (1.0 / alpha) * np.arctan(beta * np.tan(np.pi * alpha / 2))

        # Precompute trig terms
        alphaU = alpha * U
        cosU = np.cos(U)

        # Prevent division by extremely small cos(U)
        #cosU = np.where(np.abs(cosU) < 1e-14, 1e-14, cosU)

        term1 = (
            (np.sin(alphaU) + np.tan(alpha * theta0) * np.cos(alphaU))
            / cosU
        )

        inner = (
            np.cos((alpha - 1) * U)
            - np.tan(alpha * theta0) * np.sin((alpha - 1) * U)
        ) / (W * cosU)

        # Use log-exp form for numerical stability
        exponent = (1.0 - alpha) / alpha
        X = term1 * np.exp(exponent * np.log(inner))

    else:
        # -------- alpha = 1 case --------

        # Stable limit formula
        betaU = beta * U
        part1 = (np.pi / 2 + betaU) * np.tan(U)
        part2 = beta * np.log((np.pi / 2 * W * np.cos(U)) / (np.pi / 2 + betaU))

        X = (2 / np.pi) * (part1 - part2)

    return X



import numpy as np
from scipy.special import gamma
from .spectral_measure_sampler import BaseSpectralMeasureSampler
from .util import get_random_state_generator
import logging

def sample_alpha_stable_vector(
    alpha: float,
    spectral_measure: BaseSpectralMeasureSampler,
    number_of_samples: int = 1,
    shift_vector: np.ndarray = 0,
    max_number_of_convergence_terms: int = 50000,
    error: float = 0.01,
    random_state: None | int | np.random.RandomState | np.random.Generator = None
):
    random_state = get_random_state_generator(random_state)
    dimensions = spectral_measure.dimensions()
    mass = spectral_measure.mass()
    shift_vector = np.broadcast_to(shift_vector, dimensions)

    number_of_convergence_terms = int(estimate_number_of_convergence_terms(error, alpha, mass))
    if number_of_convergence_terms > max_number_of_convergence_terms:
        logging.warning(f"Estimated number of convergence terms {number_of_convergence_terms} exceeded the maximum number of convergence terms {max_number_of_convergence_terms}. Using {max_number_of_convergence_terms} convergence terms.")
        number_of_convergence_terms = max_number_of_convergence_terms

    x = np.zeros((number_of_samples, dimensions))

    cumulative_exponential = np.zeros(number_of_samples)

    for _ in range(number_of_convergence_terms):
        cumulative_exponential += random_state.exponential(scale=1.0, size=number_of_samples)

        spectral_measure_samples = spectral_measure.sample(number_of_samples, random_state)
        weights = cumulative_exponential ** (-1.0 / alpha)

        x += spectral_measure_samples * weights[:, None]

    x *= _c(alpha, mass)
    x += shift_vector

    if dimensions == 1:
        x = x.ravel()
    return x


def _c(alpha: float, mass: float):
    return (_kappa(alpha)/mass) ** (-1 / alpha)

def _kappa(alpha: float):
    if abs(alpha - 1.0) < 1e-12:
        return np.pi / 2
    return gamma(2 - alpha) * np.cos(np.pi * alpha / 2) / (1 - alpha)

def estimate_number_of_convergence_terms(error: float, alpha: float, mass: float) -> float:
    if 0 < alpha < 1:
        return estimate_number_of_convergence_terms_alpha_0_to_1(error, alpha, mass)
    elif 1 <= alpha < 2:
        return estimate_number_of_convergence_terms_alpha_1_to_2(error, alpha, mass)
    else:
        raise ValueError(f"{alpha = } must be in (0,2)")


def estimate_number_of_convergence_terms_alpha_0_to_1(mse_error: float, alpha: float, mass: float) -> float:
    assert 0 < alpha < 1 and mse_error > 0, "alpha must satisfy 0 < alpha < 1, and mse_error > 0"
    c_alpha = _c(alpha, mass)
    argument = (((1 - alpha) ** 2) * mse_error) / (2 * (alpha ** 2) * (c_alpha ** 2))
    logn = (alpha / (2 * alpha - 2)) * np.log(argument) + 1
    return np.exp(logn)


def estimate_number_of_convergence_terms_alpha_1_to_2(mse_error: float, alpha: float, mass: float) -> float:
    assert 1 <= alpha < 2 and mse_error > 0, "alpha must satisfy 1 <= alpha < 2, and mse_error > 0"
    c_alpha = _c(alpha, mass)
    log_arg = np.log(2 - alpha) + np.log(mse_error) - np.log(2 * alpha * c_alpha ** 2)
    logn = (alpha / (alpha - 2)) * log_arg
    return np.exp(logn) + 1